import sys
import platform
import time
import socket
import re
import base64

import usb.core
import usb.util
import struct
import Operations as op
import time

import numpy as np


DEVICE = None
DEVICE_TYPE = None
epBulkWriter = None
epBulkReader = None

class Monsoon(object):
    def __init__(self, *args, **kwargs):
        self._statusPacket = op.statusPacket
        self.fineThreshold = 30000
        self.auxFineThreshold = 30000
        self.mainvoltageScale = 2
        self.usbVoltageScale = 2
        self.ADCRatio = (float)(62.5 / 1e6); #Each tick of the ADC represents this much voltage
        self.factoryRes = 0.050
        self.auxFactoryResistor = 0.1
        self.padding = np.zeros(64)
        pass

    def setup_usb(self):
        global DEVICE 
        global DEVICE_TYPE
        global epBulkWriter
        global epBulkReader

        DEVICE = usb.core.find(idVendor=0x2AB9, idProduct=0x0001)
        if DEVICE is None:
            print("Error: could not find device")
            pass

        # On Linux we need to detach usb HID first
        if "Linux" == platform.system():
            try:
                DEVICE.detach_kernel_driver(0)
            except:
                pass # already unregistered    

        DEVICE.set_configuration()
        cfg = DEVICE.get_active_configuration()
        intf = cfg[(0,0)]

        epBulkWriter = usb.util.find_descriptor(
            intf,
            custom_match = \
                lambda e: \
                usb.util.endpoint_direction(e.bEndpointAddress) == \
                usb.util.ENDPOINT_OUT)
        epBulkReader = usb.util.find_descriptor(
            intf,
            custom_match = \
                lambda e: \
                usb.util.endpoint_direction(e.bEndpointAddress) == \
                usb.util.ENDPOINT_IN)

        
    def sendCommand(self,operation, value):
        """Send a USB Control Transfer"""
        try:
            value = int(value) #This was coming in as a float on OSX.
            value_array = struct.unpack("4b",struct.pack("I",value))
            operation_array = struct.unpack("4b",struct.pack("I",operation))
            wValue = struct.unpack("H",struct.pack("bb",value_array[0],value_array[1]))[0]
            wIndex = struct.unpack("H",struct.pack("bb",operation_array[0],value_array[2]))[0]
            #ctrl_transfer(bmRequestType, bmRequest, wValue, wIndex)
            DEVICE.ctrl_transfer(op.Control_Codes.USB_OUT_PACKET,op.Control_Codes.USB_SET_VALUE,wValue,wIndex,"0",5000)
        except:
            print("Control Transfer Error")
    
    def sendStopCommand(self):
        """Stop Data collection."""
        DEVICE.ctrl_transfer(op.Control_Codes.USB_IN_PACKET,op.Control_Codes.USB_REQUEST_STOP,0,0,0,5000)

    def getValue(self,operation,valueLength):
        """Read an EEPROM value"""
        operation_array = struct.unpack("4b",struct.pack("I",operation))
        wIndex = struct.unpack("H",struct.pack("bb",operation_array[0],0))[0]

        result = DEVICE.ctrl_transfer(op.Control_Codes.USB_IN_PACKET,op.Control_Codes.USB_SET_VALUE,0,wIndex,4,5000)
        result = struct.unpack("I",result)[0]
        return result

    def amps_from_raw(self,raw):
        """Translate EEPROM overcurrent protection value setting into amps."""
        raw = min(raw,1023)
        result = 8.0*(1.0-(raw/1023.0))
        return result

    def raw_from_amps(self,value):
        """Translate amp setting for overcurrent protection into EEPROM value"""
        result = 1023 * (1.0 - (amps/8.0))
        return result


    def setVout(self,value):
        """Set Vout.  Valid values from 2.01-4.55"""
        if(value > 2.00 and value < 4.56):
            vout = value * op.Conversion.FLOAT_TO_INT
            self.sendCommand(op.OpCodes.setHVMainVoltage,vout) 
        else:
            raise Exception("Invalid Voltage value")

    def setPowerupTime(self,value):
        """time in ms where powerupcurrentlimit applies, vs runtimecurrentlimit"""
        self.sendCommand(op.OpCodes.setPowerupTime,value)

    def setPowerUpCurrentLimit(self, value):
        """Sets power up current limit.  Valid values are 0-8 Amps"""
        value = self.raw_from_amps(value)
        self.sendCommand(op.OpCodes.SetHVPowerUpCurrentLimit,value)

    def setRunTimeCurrentLimit(self, value):
        """Sets runtime current limit.  Valid values are 0-8 amps."""
        value = self.raw_from_amps(value)
        self.sendCommand(op.OpCodes.SetHVRunCurrentLimit,value)

    def setMainFineScale(self, value):
        """Current scaling for 0-45 mA range"""
        self.sendCommand(op.OpCodes.setMainFineScale,value)

    def setMainCoarseScale(self,value):
        """Current scaling for main channel, 45 mA - 4.5A range"""
        self.sendCommand(op.OpCodes.setMainCoarseScale,value)

    def setUSBFineScale(self, value):
        """Current scaling for USB channel, 0-45 mA range"""
        self.sendCommand(op.OpCodes.setUSBFineScale,value)

    def setUSBCoarseScale(self,value):
        """Current scaling for USB channel, 45 mA - 1A range"""
        self.sendCommand(op.OpCodes.setUSBCoarseScale,value)

    def setAuxFineScale(self,value):
        """Current scaling for Aux Channel, 0-10 mA range"""
        self.sendCommand(op.OpCodes.setAuxFineScale,value)

    def setAuxCoarseScale(self,value):
        """Current scaling for Aux channel, 10 mA - 1A range"""
        self.sendCommand(op.OpCodes.setAuxCoarseScale,value)

    def setMainFineZeroOffset(self, value):
        """Zero offset, not used in LVPM"""
        self.sendCommand(op.OpCodes.SetMainFineZeroOffset,value)

    def setMainCoarseZeroOffset(self,value):
        """Zero offset, not used in LVPM"""
        self.sendCommand(op.OpCodes.SetMainCoarseZeroOffset,value)

    def setUSBFineZeroOffset(self, value):
        """Zero offset, not used in LVPM"""
        self.sendCommand(op.OpCodes.SetUSBFineZeroOffset,value)

    def setUSBCoarseZeroOffset(self,value):
        """Zero offset, not used in LVPM"""
        self.sendCommand(op.OpCodes.SetUSBCoarseZeroOffset,value)

    def setUSBPassthroughMode(self, USBPassthroughCode):
        """USB Passthrough mode.  0 = off, 1 = on, 2 = auto"""
        self.sendCommand(op.OpCodes.setUsbPassthroughMode,USBPassthroughCode)

    def setVoltageChannel(self, VoltageChannelCode):
        """Sets voltage measurement channel.  0 = Main & USB, 1 = Main & Aux"""
        self.sendCommand(op.OpCodes.setHVVoltageChannel,value)

    def getVoltageChannel(self):
        """0 = Main & USB, 1 = Main & Aux"""
        return(self.getValue(op.OpCodes.setHVVoltageChannel,1))
    def getSerialNumber(self):
        """Get the device serial number"""
        serialNumber = self.getValue(op.OpCodes.getSerialNumber,2)
        return serialNumber

    def fillStatusPacket(self):
        """Get all calibration information from the device EEPROM"""
        self._statusPacket.mainFineResistorOffset = float(self.getValue(op.OpCodes.setMainFineResistorOffset,1))
        mainFineResistor = 0.05 + self._statusPacket.mainFineResistorOffset * 0.0001
        self._statusPacket.mainCoarseResistorOffset = float(self.getValue(op.OpCodes.setMainCoarseResistorOffset,1))
        mainCoarseResistor = 0.05 + self._statusPacket.mainCoarseResistorOffset * 0.0001
        self._statusPacket.usbFineResistorOffset = float(self.getValue(op.OpCodes.setUsbFineResistorOffset,1))
        usbFineResistor = 0.05 + self._statusPacket.usbFineResistorOffset * 0.0001
        self._statusPacket.usbCoarseResistorOffset = float(self.getValue(op.OpCodes.setUsbCoarseResistorOffset,1))
        usbCoarseResistor = 0.05 + self._statusPacket.usbCoarseResistorOffset * 0.0001
        self._statusPacket.auxFineResistorOffset = float(self.getValue(op.OpCodes.setAuxFineResistorOffset,1))
        auxFineResistor = 0.1 + self._statusPacket.auxFineResistorOffset * 0.0001
        self._statusPacket.auxCoarseResistorOffset = float(self.getValue(op.OpCodes.setAuxCoarseResistorOffset,1))
        auxCoarseResistor = 0.1 + self._statusPacket.auxCoarseResistorOffset * 0.0001

        
        self._statusPacket.mainFineScale = 35946.0 * (self.factoryRes / mainFineResistor)
        self._statusPacket.mainCoarseScale = 3103.4 * (self.factoryRes / mainCoarseResistor)
        self._statusPacket.usbFineScale = 35946.0 * (self.factoryRes / usbFineResistor)
        self._statusPacket.usbCoarseScale = 3103.4 * (self.factoryRes / usbCoarseResistor)
        self._statusPacket.auxFineScale = 0.05049 * (self.auxFactoryResistor / auxFineResistor)
        self._statusPacket.auxCoarseScale = 1.289 * (self.auxFactoryResistor / auxCoarseResistor)

        self._statusPacket.mainFineZeroOffset = 0
        self._statusPacket.mainCoarseZeroOffset = 0
        self._statusPacket.usbFineZeroOffset = 0
        self._statusPacket.usbCoarseZeroOffset = 0


        #ctrl_transfer(bmRequestType, bmRequest, wValue, wIndex)
    def Start(self,calTime, maxTime):
        """Start sampling.  calTime = time in ms between collecting calibration packets.  maxTime = time in ms the PM will continue sampling.
        Use maxTime = 0xFFFFFFFF for indefinite sampling"""
        self.fillStatusPacket()
        value_array = struct.unpack("4B",struct.pack("I",calTime))
        maxtime_array = struct.unpack("4B",struct.pack("I",maxTime))
        wValue = struct.unpack("H",struct.pack("BB",value_array[0],value_array[1]))[0]
        wIndex = struct.unpack("H",struct.pack("BB",op.Control_Codes.USB_REQUEST_START,0))[0]
        DEVICE.ctrl_transfer(op.Control_Codes.USB_OUT_PACKET,op.Control_Codes.USB_REQUEST_START,wValue,wIndex,maxtime_array,1000)

    def BulkRead(self):
        """Read a sample packet"""
        return(DEVICE.read(0x81,64,timeout=0))

    def swizzlePacket(self, packet):
        """Byte order for 2-byte values is swapped for the majority of the measurement packet.  This fixes that."""
        length = len(packet)
        packet = np.array(packet)
        evenBytes = packet[4::2]
        oddBytes = packet[5::2]
        swizzledBytes = np.insert(evenBytes,np.arange(len(oddBytes)),oddBytes)
        swizzledPacket = np.hstack([packet[0:4],swizzledBytes[:],self.padding[0:(58-length)]])
        swizzledPacket = np.array(swizzledPacket).astype('b')
        packetLength = len(swizzledPacket)
        rawBytes = struct.pack('b'*packetLength,*swizzledPacket)
        measurements = struct.unpack("HBBhhhhhhhHBBhhhhhhhHBBhhhhhhhHBB",rawBytes)
        measurements = list(measurements)
        return measurements





